@@ -1,17 +1,27 @@
 //! Port of https://github.com/styled-components/babel-plugin-styled-components/blob/a20c3033508677695953e7a434de4746168eeb4e/src/visitors/transpileCssProp.js
 
+use std::{borrow::Cow, collections::HashMap};
+
 use inflector::Inflector;
 use once_cell::sync::Lazy;
 use regex::Regex;
 use swc_atoms::{js_word, JsWord};
-use swc_common::{util::take::Take, Spanned, DUMMY_SP};
+use swc_common::{
+    collections::{AHashMap, AHashSet},
+    util::take::Take,
+    Spanned, DUMMY_SP,
+};
 use swc_ecmascript::{
     ast::*,
-    utils::{prepend, private_ident, quote_ident, quote_str, ExprExt, ExprFactory},
+    utils::{
+        ident::IdentLike, prepend, private_ident, quote_ident, quote_str, ExprExt, ExprFactory, Id,
+    },
     visit::{as_folder, noop_visit_mut_type, Fold, VisitMut, VisitMutWith},
 };
 
-use crate::utils::{get_prop_key_as_expr, get_prop_name};
+use crate::utils::{get_prop_key_as_expr, get_prop_name, get_prop_name2};
+
+use super::top_level_binding_collector::collect_top_level_decls;
 
 static TAG_NAME_REGEX: Lazy<Regex> =
     Lazy::new(|| Regex::new("^[a-z][a-z\\d]*(\\-[a-z][a-z\\d]*)?$").unwrap());
@@ -24,8 +34,25 @@ pub fn transpile_css_prop() -> impl Fold + VisitMut {
 struct TranspileCssProp {
     import_name: Option<Ident>,
     injected_nodes: Vec<Stmt>,
+    interleaved_injections: AHashMap<Id, Vec<Stmt>>,
 
     identifier_idx: usize,
+    styled_idx: HashMap<JsWord, usize>,
+    top_level_decls: Option<AHashSet<Id>>,
+}
+
+impl TranspileCssProp {
+    fn next_styled_idx(&mut self, key: JsWord) -> usize {
+        let idx = self.styled_idx.entry(key).or_insert(0);
+        *idx += 1;
+        *idx
+    }
+    fn is_top_level_ident(&mut self, ident: &Ident) -> bool {
+        self.top_level_decls
+            .as_ref()
+            .map(|decls| decls.contains(&ident.to_id()))
+            .unwrap_or(false)
+    }
 }
 
 impl VisitMut for TranspileCssProp {
@@ -48,27 +75,34 @@ impl VisitMut for TranspileCssProp {
                         .get_or_insert_with(|| private_ident!("_styled"))
                         .clone();
 
-                    let name = get_name(&elem.opening.name);
-                    let id_sym = name.to_class_case();
+                    let name = get_name_ident(&elem.opening.name);
+                    let id_sym = name.sym.to_class_case();
 
-                    let id: Ident =
-                        private_ident!(elem.opening.name.span(), format!("_Styled{}", id_sym));
+                    // Match the original plugin's behavior.
+                    let id_sym = id_sym.trim_end_matches(char::is_numeric);
 
-                    let (styled, inject_after) = if TAG_NAME_REGEX.is_match(&name) {
+                    let id_sym = JsWord::from(id_sym);
+                    let styled_idx = self.next_styled_idx(id_sym.clone());
+                    let id = quote_ident!(
+                        elem.opening.name.span(),
+                        append_if_gt_one(&format!("_Styled{}", id_sym), styled_idx)
+                    );
+
+                    let (styled, inject_after) = if TAG_NAME_REGEX.is_match(&name.sym) {
                         (
                             (Expr::Call(CallExpr {
                                 span: DUMMY_SP,
                                 callee: import_name.as_callee(),
                                 args: vec![Lit::Str(Str {
                                     span: DUMMY_SP,
-                                    value: name.into(),
+                                    value: name.sym.into(),
                                     has_escape: false,
                                     kind: Default::default(),
                                 })
                                 .as_arg()],
                                 type_args: Default::default(),
                             })),
-                            None::<()>,
+                            None::<Ident>,
                         )
                     } else {
                         let name_expr = get_name_expr(&elem.opening.name);
@@ -80,7 +114,11 @@ impl VisitMut for TranspileCssProp {
                                 args: vec![name_expr.as_arg()],
                                 type_args: Default::default(),
                             }),
-                            None,
+                            if self.is_top_level_ident(&name) {
+                                Some(name)
+                            } else {
+                                None
+                            },
                         )
                     };
 
@@ -193,34 +231,45 @@ impl VisitMut for TranspileCssProp {
                                 .fold(vec![], |mut acc, mut expr| {
                                     if expr.is_fn_expr() || expr.is_arrow() {
                                         acc.push(expr);
-                                    } else {
-                                        let identifier =
-                                            get_local_identifier(&mut self.identifier_idx, &expr);
-                                        let p = quote_ident!("P");
-                                        extra_attrs.push(JSXAttrOrSpread::JSXAttr(JSXAttr {
-                                            span: DUMMY_SP,
-                                            name: JSXAttrName::Ident(identifier.clone()),
-                                            value: Some(JSXAttrValue::JSXExprContainer(
-                                                JSXExprContainer {
-                                                    span: DUMMY_SP,
-                                                    expr: JSXExpr::Expr(expr.take()),
-                                                },
-                                            )),
-                                        }));
-
-                                        acc.push(Box::new(Expr::Arrow(ArrowExpr {
-                                            span: DUMMY_SP,
-                                            params: vec![Pat::Ident(p.clone().into())],
-                                            body: BlockStmtOrExpr::Expr(Box::new(
-                                                p.make_member(identifier),
-                                            )),
-                                            is_async: false,
-                                            is_generator: false,
-                                            type_params: Default::default(),
-                                            return_type: Default::default(),
-                                        })))
+                                        return acc;
+                                    } else if let Some(root) = trace_root_value(&mut *expr) {
+                                        let direct_access = match root {
+                                            Expr::Lit(_) => true,
+                                            Expr::Ident(id) if self.is_top_level_ident(id) => true,
+                                            _ => false,
+                                        };
+                                        if direct_access {
+                                            acc.push(expr);
+                                            return acc;
+                                        }
                                     }
 
+                                    let identifier =
+                                        get_local_identifier(&mut self.identifier_idx, &expr);
+                                    let p = quote_ident!("p");
+                                    extra_attrs.push(JSXAttrOrSpread::JSXAttr(JSXAttr {
+                                        span: DUMMY_SP,
+                                        name: JSXAttrName::Ident(identifier.clone()),
+                                        value: Some(JSXAttrValue::JSXExprContainer(
+                                            JSXExprContainer {
+                                                span: DUMMY_SP,
+                                                expr: JSXExpr::Expr(expr.take()),
+                                            },
+                                        )),
+                                    }));
+
+                                    acc.push(Box::new(Expr::Arrow(ArrowExpr {
+                                        span: DUMMY_SP,
+                                        params: vec![Pat::Ident(p.clone().into())],
+                                        body: BlockStmtOrExpr::Expr(Box::new(
+                                            p.make_member(identifier),
+                                        )),
+                                        is_async: false,
+                                        is_generator: false,
+                                        type_params: Default::default(),
+                                        return_type: Default::default(),
+                                    })));
+
                                     acc
                                 });
 
@@ -246,15 +295,22 @@ impl VisitMut for TranspileCssProp {
                         }),
                         definite: false,
                     };
+                    let stmt = Stmt::Decl(Decl::Var(VarDecl {
+                        span: DUMMY_SP,
+                        kind: VarDeclKind::Var,
+                        declare: false,
+                        decls: vec![var],
+                    }));
                     match inject_after {
-                        Some(_injector) => todo!("Use injector"),
+                        Some(injector) => {
+                            let id = injector.to_id();
+                            self.interleaved_injections
+                                .entry(id)
+                                .or_default()
+                                .push(stmt);
+                        }
                         None => {
-                            self.injected_nodes.push(Stmt::Decl(Decl::Var(VarDecl {
-                                span: DUMMY_SP,
-                                kind: VarDeclKind::Var,
-                                declare: false,
-                                decls: vec![var],
-                            })));
+                            self.injected_nodes.push(stmt);
                         }
                     }
                 }
@@ -285,7 +341,9 @@ impl VisitMut for TranspileCssProp {
 
     fn visit_mut_module(&mut self, n: &mut Module) {
         // TODO: Skip if there are no css prop usage
+        self.top_level_decls = Some(collect_top_level_decls(n));
         n.visit_mut_children_with(self);
+        self.top_level_decls = None;
 
         if let Some(import_name) = self.import_name.take() {
             let specifier = ImportSpecifier::Default(ImportDefaultSpecifier {
@@ -309,6 +367,37 @@ impl VisitMut for TranspileCssProp {
             );
         }
 
+        let mut serialized_body: Vec<ModuleItem> = vec![];
+        let body = std::mem::replace(&mut n.body, vec![]);
+        for item in body {
+            serialized_body.push(item.clone());
+            match &item {
+                ModuleItem::Stmt(Stmt::Decl(Decl::Var(vd))) => {
+                    for decl in &vd.decls {
+                        if let Pat::Ident(ident) = &decl.name {
+                            let id = ident.to_id();
+                            let stmts = self.interleaved_injections.remove(&id);
+                            if let Some(stmts) = stmts {
+                                serialized_body
+                                    .extend(stmts.into_iter().rev().map(ModuleItem::Stmt));
+                            }
+                        }
+                    }
+                }
+                _ => {}
+            }
+        }
+        n.body = serialized_body;
+
+        let mut remaining = std::mem::replace(&mut self.interleaved_injections, Default::default())
+            .into_iter()
+            .collect::<Vec<_>>();
+        remaining.sort_by_key(|x| x.0.clone());
+
+        remaining
+            .into_iter()
+            .for_each(|(_, stmts)| n.body.extend(stmts.into_iter().map(ModuleItem::Stmt)));
+
         n.body
             .extend(self.injected_nodes.take().into_iter().map(ModuleItem::Stmt));
     }
@@ -436,13 +525,10 @@ impl PropertyReducer<'_> {
                         })),
                     }));
 
-                    let key = get_prop_key_as_expr(&prop);
+                    let key = get_prop_name2(&prop);
 
                     acc.push(PropOrSpread::Prop(Box::new(Prop::KeyValue(KeyValueProp {
-                        key: PropName::Computed(ComputedPropName {
-                            span: DUMMY_SP,
-                            expr: Box::new(key.into_owned()),
-                        }),
+                        key,
                         value: Box::new(self.p.clone().make_member(identifier)),
                     }))));
                 } else {
@@ -500,21 +586,31 @@ fn set_key_of_prop(prop: &mut Prop, key: Box<Expr>) {
 fn get_local_identifier(idx: &mut usize, expr: &Expr) -> Ident {
     *idx += 1;
 
-    let identifier = quote_ident!(expr.span(), format!("$_css{}", *idx));
+    let identifier = quote_ident!(expr.span(), append_if_gt_one("$_css", *idx));
 
     // TODO: Unique identifier
 
     identifier
 }
 
-fn get_name(el: &JSXElementName) -> JsWord {
+fn append_if_gt_one(s: &str, suffix: usize) -> Cow<str> {
+    if suffix > 1 {
+        Cow::Owned(format!("{}{}", s, suffix))
+    } else {
+        Cow::Borrowed(s)
+    }
+}
+
+fn get_name_ident(el: &JSXElementName) -> Ident {
     match el {
-        JSXElementName::Ident(v) => v.sym.clone(),
-        JSXElementName::JSXMemberExpr(e) => {
-            format!("{}_{}", get_name_of_jsx_obj(&e.obj), e.prop.sym).into()
-        }
+        JSXElementName::Ident(v) => v.clone(),
+        JSXElementName::JSXMemberExpr(e) => Ident {
+            sym: format!("{}_{}", get_name_of_jsx_obj(&e.obj), e.prop.sym).into(),
+            span: e.prop.span,
+            optional: false,
+        },
         _ => {
-            unimplemented!("get_name for namespaced jsx element")
+            unimplemented!("get_name_ident for namespaced jsx element")
         }
     }
 }
@@ -527,3 +623,16 @@ fn get_name_of_jsx_obj(el: &JSXObject) -> JsWord {
         }
     }
 }
+
+fn trace_root_value(e: &mut Expr) -> Option<&mut Expr> {
+    match e {
+        Expr::Member(e) => trace_root_value(&mut e.obj),
+        Expr::Call(e) => match &mut e.callee {
+            Callee::Expr(e) => trace_root_value(&mut **e),
+            _ => None,
+        },
+        Expr::Ident(_) => Some(e),
+        Expr::Lit(_) => Some(e),
+        _ => None,
+    }
+}
